import base64
import os
import random
import logging
import sys
import asyncio
import json

from google.protobuf.json_format import MessageToDict

from typing import Annotated

from fastapi import FastAPI, File, HTTPException, Header

from proto_ems_server.device_auth_pb2 import Context


# Constants
ERR_PERCENT = float(os.getenv("ERR_PERCENT", "0.0034"))
ERR_5XX_PERCENT = float(os.getenv("ERR_4XX_PERCENT", "0.0025"))
LOW_2XX_LATENCY = int(os.getenv("LOW_2XX_LATENCY", "180"))
HIGH_2XX_LATENCY = int(os.getenv("HIGH_2XX_LATENCY", "240"))
SOCKET_TIMEOUT_PERCENT = float(os.getenv("SOCKET_TIMEOUT_PERCENT", "0.11"))
SOCKET_TIMEOUT_LATENCY = int(os.getenv("SOCKET_TIMEOUT_LATENCY", "90"))
EOF_EXCEPTION_LATENCY = int(os.getenv("EOF_EXCEPTION_LATENCY", "23"))
EMULATE_LATENCIES = os.getenv("EMULATE_LATENCIES", "true").lower() == "true"

app = FastAPI()

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)
stream_handler = logging.StreamHandler(sys.stdout)
log_formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(name)s: %(message)s")
stream_handler.setFormatter(log_formatter)
log.addHandler(stream_handler)


@app.get("/")
def read_root():
    """
    Default route
    :return: {'Hello': 'World'}
    """
    return {"Hello": "World"}


@app.post("/device-auth-mock", status_code=204)
async def device_auth_mock(device_context: Annotated[str | None, Header()] = None):
    """
    Mock endpoint for ems-device-auth.
    :param device_context: The Device-Context header for the request
    :return: None
    """
    if device_context:
        device_id = _decode_device_id(device_context)
    else:
        device_id = "not-found"
    if EMULATE_LATENCIES:
        return await _emulate_latency(device_id)
    else:
        return


@app.post("/hello-mock")
def hello_mock(req_file: bytes = File(...)):
    """
    Mock endpoint for hello requests. Meant to simply echo with the size of the request.
    :param req_file: Packed msgpack hello request
    :return: List of mock hints
    """
    return [{"mock_hint": len(req_file)}]


@app.post("/th-event-mock")
def th_event_mock(device_context: Annotated[str | None, Header()] = None, th_event: bytes = File(...)):
    """
    Mock endpoint for threathunter events. Meant to serve as a sink and not to pass events downstream
    :param device_context: The Device-Context header for the request
    :param th_event: The event file
    :return: None
    """
    if device_context:
        device_id = _decode_device_id(device_context)
    else:
        device_id = "not-found"
    log.info(f"Received threathunter event for device {device_id} with filesize: {len(th_event)}")
    return


async def _emulate_latency(device_id: str):
    """
    Adds latency and emulates 4xx response frequency for ems-device-auth
    :param device_id: The device for which this request is made
    :return: None
    """
    error_probability = random.random()
    if error_probability < ERR_PERCENT:
        code_probability = random.random()
        if code_probability < ERR_5XX_PERCENT:
            pass
        else:
            timeout_probability = random.random()
            if timeout_probability < SOCKET_TIMEOUT_PERCENT:
                await asyncio.sleep(SOCKET_TIMEOUT_LATENCY)
                log.warning(f"Returning 400: SocketTimeoutException for device {device_id}")
                raise HTTPException(status_code=400, detail="SocketTimeoutException")
            else:
                await asyncio.sleep(EOF_EXCEPTION_LATENCY)
                log.warning(f"Returning 400: EOFException for device {device_id}")
                raise HTTPException(status_code=400, detail="EOFException")
        return
    else:
        latency = random.randint(LOW_2XX_LATENCY, HIGH_2XX_LATENCY)
        await asyncio.sleep(latency / 1000)
        log.info(f"Returning 204 response for {device_id}")
        return


def _decode_device_id(device_context: str) -> str:
    """
    Decode Device-Context header into Context protobuf message and get device id from that message
    :param device_context: Base64 encoded serialized string representation of Context protobuf message
    :return: device id from Device-Context header
    """
    context = Context()
    decoded_context = base64.b64decode(device_context)
    context.ParseFromString(decoded_context)
    context_dict = MessageToDict(context)
    return context_dict.get('device', {}).get('id')
